@@ -1,5 +1,5 @@
 /*
-    Copyright 2008, Somsak Sriprayoonsakul <somsaks@gmail.com>
+    Copyright 2008-2010,  Somsak Sriprayoonsakul  <somsaks@gmail.com>
     
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -53,6 +53,7 @@
 
 #include "mb_net.h"
 #include "mb_util.h"
+#include "mb_cache_util.h"
 
 #ifdef _WIN32
 #	include <win32dep.h>
@@ -68,8 +69,12 @@
 #define TW_ACCT_LAST_MSG_ID "twitter_last_msg_id"
 #define TW_ACCT_SENT_MSG_IDS "twitter_sent_msg_ids"
 
-//static const char twitter_host[] = "twitter.com";
-static gint twitter_port = 443;
+const char * mb_auth_types_str[] = {
+		"mb_oauth",
+		"mb_xauth",
+		"mb_http_basicauth",
+};
+
 static const char twitter_fixed_headers[] = "User-Agent:" TW_AGENT "\r\n" \
 "Accept: */*\r\n" \
 "X-Twitter-Client: " TW_AGENT_SOURCE "\r\n" \
@@ -80,11 +85,90 @@ static const char twitter_fixed_headers[] = "User-Agent:" TW_AGENT "\r\n" \
 
 PurplePlugin * twitgin_plugin = NULL;
 
+static MbConnData * twitter_init_connection(MbAccount * ma, gint type, const char * path, MbHandlerFunc handler);
+static gint twitter_oauth_prepare(MbConnData * conn_data, gpointer data, const char * error);
+void twitter_request_access(MbAccount * ma);
+gint twitter_request_authorize(MbAccount * ma, MbConnData * data, gpointer user_data);
+void twitter_request_authorize_ok_cb(MbAccount * ma, const char * pin);
+gint twitter_oauth_request_finish(MbAccount * ma, MbConnData * data, gpointer user_data);
+void twitter_verify_account(MbAccount * ma, gpointer data);
+gint twitter_verify_authen(MbConnData * conn_data, gpointer data, const char * error);
+
+/**
+ * Convenient function to initialize new connection and set necessary value
+ */
+//
+static MbConnData * twitter_init_connection(MbAccount * ma, gint type, const char * path, MbHandlerFunc handler)
+{
+	MbConnData * conn_data = NULL;
+	gboolean use_https = purple_account_get_bool(ma->account, mc_name(TC_USE_HTTPS), mc_def_bool(TC_USE_HTTPS));
+	gint retry = purple_account_get_int(ma->account, mc_name(TC_GLOBAL_RETRY), mc_def_int(TC_GLOBAL_RETRY));
+	gint port;
+	gchar * user_name = NULL, * host = NULL;
+	const char * password;
+
+	if(use_https) {
+		port = TW_HTTPS_PORT;
+	} else {
+		port = TW_HTTP_PORT;
+	}
+
+	twitter_get_user_host(ma, &user_name, &host);
+	password = purple_account_get_password(ma->account);
+
+	conn_data = mb_conn_data_new(ma, host, port, handler, use_https);
+	mb_conn_data_set_retry(conn_data, retry);
+
+	conn_data->request->type = type;
+	conn_data->request->port = port;
+
+	mb_http_data_set_host(conn_data->request, host);
+	mb_http_data_set_path(conn_data->request, path);
+	// XXX: Use global here -> twitter_fixed_headers
+	mb_http_data_set_fixed_headers(conn_data->request, twitter_fixed_headers);
+	mb_http_data_set_header(conn_data->request, "Host", host);
+
+
+	switch(ma->auth_type) {
+		case MB_OAUTH :
+		case MB_XAUTH :
+			// attach oauth header with this connection
+			if(ma->oauth.oauth_token && ma->oauth.oauth_secret) {
+				conn_data->prepare_handler = twitter_oauth_prepare;
+				conn_data->prepare_handler_data = ma;
+			}
+			break;
+		default :
+			// basic auth is default
+			mb_http_data_set_basicauth(conn_data->request, user_name, password);
+			break;
+	}
+	if(user_name) g_free(user_name);
+	if(host) g_free(host);
+
+	return conn_data;
+}
+
+static gint twitter_oauth_prepare(MbConnData * conn_data, gpointer data, const char * error) {
+	MbAccount * ma = (MbAccount *)data;
+	gchar * full_url;
+
+	// error is always NULL here
+	full_url = mb_conn_url_unparse(conn_data);
+	if(conn_data->retry <= 0) {
+		mb_oauth_set_http_data(&ma->oauth, conn_data->request, full_url, conn_data->request->type);
+	} else {
+		mb_oauth_reset_nonce(&ma->oauth, conn_data->request, full_url, conn_data->request->type);
+	}
+	g_free(full_url);
+	return 0;
+}
+
 static TwitterBuddy * twitter_new_buddy()
 {
 	TwitterBuddy * buddy = g_new(TwitterBuddy, 1);
 	
-	buddy->ta = NULL;
+	buddy->ma = NULL;
 	buddy->buddy = NULL;
 	buddy->uid = -1;
 	buddy->name = NULL;
@@ -102,6 +186,7 @@ TwitterTimeLineReq * twitter_new_tlr(const char * path, const char * name, int i
 	tlr->count = count;
 	tlr->timeline_id = id;
 	tlr->use_since_id = TRUE;
+	tlr->screen_name = NULL;
 	if(sys_msg) {
 		tlr->sys_msg = g_strdup(sys_msg);
 	} else {
@@ -155,59 +240,62 @@ void twitter_buddy_free(PurpleBuddy * buddy)
 // Privacy mode skip fetching messages during unavailable state
 gboolean twitter_skip_fetching_messages(PurpleAccount * acct) 
 {
-	gboolean privacy_mode = purple_account_get_bool(acct, tc_name(TC_PRIVACY), tc_def_bool(TC_PRIVACY));
+	MbAccount * ma = (MbAccount *)acct->gc->proto_data;
+	gboolean privacy_mode = purple_account_get_bool(acct, mc_name(TC_PRIVACY), mc_def_bool(TC_PRIVACY));
 	gboolean available = purple_status_is_available(purple_account_get_active_status(acct));
 
 	if(privacy_mode && !available) {
 		purple_debug_info(DBGID, "Unavailable, skipping fetching due privacy mode\n");
 		return TRUE;
 	}
-	else {
-		return FALSE;
+	if(!purple_privacy_check(acct, mc_def(TC_FRIENDS_USER))) {
+		purple_debug_info(DBGID, "Privacy block, skipping fetching due privacy mode\n");
+		return TRUE;
 	}
+	return FALSE;
 }
 
 // Function to fetch first batch of new message
-void twitter_fetch_first_new_messages(TwitterAccount * ta)
+void twitter_fetch_first_new_messages(MbAccount * ma)
 {
-	if(twitter_skip_fetching_messages(ta->account)) {
-		return;
-	}
-
 	TwitterTimeLineReq * tlr;
 	const gchar * tl_path;
 	int count;
 	
+	if(twitter_skip_fetching_messages(ma->account)) {
+		return;
+	}
+
 	purple_debug_info(DBGID, "%s called\n", __FUNCTION__);
-	tl_path = purple_account_get_string(ta->account, tc_name(TC_FRIENDS_TIMELINE), tc_def(TC_FRIENDS_TIMELINE));
-	count = purple_account_get_int(ta->account, tc_name(TC_INITIAL_TWEET), tc_def_int(TC_INITIAL_TWEET));
+	tl_path = purple_account_get_string(ma->account, mc_name(TC_FRIENDS_TIMELINE), mc_def(TC_FRIENDS_TIMELINE));
+	count = purple_account_get_int(ma->account, mc_name(TC_INITIAL_TWEET), mc_def_int(TC_INITIAL_TWEET));
 	purple_debug_info(DBGID, "count = %d\n", count);
-	tlr = twitter_new_tlr(tl_path, tc_def(TC_FRIENDS_USER), TL_FRIENDS, count, NULL);
-	twitter_fetch_new_messages(ta, tlr);
+	tlr = twitter_new_tlr(tl_path, mc_def(TC_FRIENDS_USER), TL_FRIENDS, count, NULL);
+	twitter_fetch_new_messages(ma, tlr);
 }
 
 // Function to fetch all new messages periodically
 gboolean twitter_fetch_all_new_messages(gpointer data)
 {
-	TwitterAccount * ta = data;
+	MbAccount * ma = data;
 	TwitterTimeLineReq * tlr = NULL;
 	gint i;
 	const gchar * tl_path;
 
-	if(twitter_skip_fetching_messages(ta->account)) {
+	if(twitter_skip_fetching_messages(ma->account)) {
 		return TRUE;
 	}
 	
 	for(i = TC_FRIENDS_TIMELINE; i <= TC_USER_TIMELINE; i+=2) {
 		//FIXME: i + 1 is not a very good strategy here
-		if(!purple_find_buddy(ta->account, tc_def(i + 1))) {
+		if(!purple_find_buddy(ma->account, mc_def(i + 1))) {
 			purple_debug_info(DBGID, "skipping %s\n", tlr->name);
 			continue;
 		}
-		tl_path = purple_account_get_string(ta->account, tc_name(i), tc_def(i));
-		tlr = twitter_new_tlr(tl_path, tc_def(i + 1), i, TW_STATUS_COUNT_MAX, NULL);
+		tl_path = purple_account_get_string(ma->account, mc_name(i), mc_def(i));
+		tlr = twitter_new_tlr(tl_path, mc_def(i + 1), i, TW_STATUS_COUNT_MAX, NULL);
 		purple_debug_info(DBGID, "fetching updates from %s to %s\n", tlr->path, tlr->name);
-		twitter_fetch_new_messages(ta, tlr);
+		twitter_fetch_new_messages(ma, tlr);
 	}
 	return TRUE;
 }
@@ -247,10 +335,10 @@ char * twitter_decode_error(const char * data)
 GList * twitter_decode_messages(const char * data, time_t * last_msg_time)
 {
 	GList * retval = NULL;
-	xmlnode * top = NULL, *id_node, *time_node, *status, * text, * user, * user_name, * image_url;
-	gchar * from, * msg_txt, * avatar_url, *xml_str = NULL;
+	xmlnode * top = NULL, *id_node, *time_node, *status, * text, * user, * user_name, * image_url, * user_is_protected;
+	gchar * from, * msg_txt, * avatar_url = NULL, *xml_str = NULL, * is_protected = NULL;
 	TwitterMsg * cur_msg = NULL;
-	unsigned long long cur_id;
+	mb_status_t cur_id;
 	time_t msg_time_t;
 
 	purple_debug_info(DBGID, "%s called\n", __FUNCTION__);
@@ -284,7 +372,7 @@ GList * twitter_decode_messages(const char * data, time_t * last_msg_time)
 			xml_str = xmlnode_get_data_unescaped(time_node);
 		}
 		purple_debug_info(DBGID, "msg time = %s\n", xml_str);
-		msg_time_t = mb_mktime(xml_str) - timezone;
+		msg_time_t = mb_mktime(xml_str);
 		if( (*last_msg_time) < msg_time_t) {
 			(*last_msg_time) = msg_time_t;
 		}
@@ -304,9 +392,13 @@ GList * twitter_decode_messages(const char * data, time_t * last_msg_time)
 				from = xmlnode_get_data(user_name);
 			}
 			image_url = xmlnode_get_child(user, "profile_image_url");
-			if(user_name) {
+			if(image_url) {
 				avatar_url = xmlnode_get_data(image_url);
 			}
+			user_is_protected = xmlnode_get_child(user, "protected");
+			if(user_is_protected) {
+				is_protected = xmlnode_get_data(user_is_protected);
+			}
 		}
 
 		if(from && msg_txt) {
@@ -314,9 +406,15 @@ GList * twitter_decode_messages(const char * data, time_t * last_msg_time)
 			
 			purple_debug_info(DBGID, "from = %s, msg = %s\n", from, msg_txt);
 			cur_msg->id = cur_id;
-			cur_msg->from = from; //< actually we don't need this for now
+			cur_msg->from = from;
 			cur_msg->avatar_url = avatar_url; //< actually we don't need this for now
 			cur_msg->msg_time = msg_time_t;
+			if(is_protected && (strcmp(is_protected, "false") == 0) ) {
+				cur_msg->is_protected = FALSE;
+				g_free(is_protected);
+			} else {
+				cur_msg->is_protected = TRUE;
+			}
 			cur_msg->flag = 0;
 			/*
 			if(skip) {
@@ -335,7 +433,7 @@ GList * twitter_decode_messages(const char * data, time_t * last_msg_time)
 	return retval;
 }
 
-gint twitter_fetch_new_messages_handler(MbConnData * conn_data, gpointer data)
+gint twitter_fetch_new_messages_handler(MbConnData * conn_data, gpointer data, const char * error)
 {
 	MbAccount * ma = conn_data->ma;
 	const gchar * username;
@@ -350,6 +448,11 @@ gint twitter_fetch_new_messages_handler(MbConnData * conn_data, gpointer data)
 	purple_debug_info(DBGID, "%s called\n", __FUNCTION__);
 	purple_debug_info(DBGID, "received result from %s\n", tlr->path);
 	
+	if(error) {
+		// we don't want to handle network error
+		return 0;
+	}
+
 	username = (const gchar *)purple_account_get_username(ma->account);
 	
 	if(response->status == HTTP_MOVED_TEMPORARILY) {
@@ -360,16 +463,15 @@ gint twitter_fetch_new_messages_handler(MbConnData * conn_data, gpointer data)
 	}
 	if(response->status != HTTP_OK) {
 		twitter_free_tlr(tlr);
-		if(response->status == HTTP_BAD_REQUEST) {
+		if((response->status == HTTP_BAD_REQUEST) || (response->status == HTTP_UNAUTHORIZE)) {
 			// rate limit exceed?
 			if(response->content_len > 0) {
 				gchar * error_str = NULL;
 
 				error_str = twitter_decode_error(response->content->str);
 				if(ma->gc != NULL) {
-					purple_connection_set_state(ma->gc, PURPLE_DISCONNECTED);
-					ma->state = PURPLE_DISCONNECTED;
-					purple_connection_error(ma->gc, error_str);
+					// XXX: Do we need to use conn_error here? Since we will return 0 below...
+					mb_conn_error(conn_data, PURPLE_CONNECTION_ERROR_OTHER_ERROR, error_str);
 				}
 				g_free(error_str);
 			}
@@ -393,7 +495,7 @@ gint twitter_fetch_new_messages_handler(MbConnData * conn_data, gpointer data)
 	
 	// reverse the list and append it
 	// only if id > last_msg_id
-	hide_myself = purple_account_get_bool(ma->account, tc_name(TC_HIDE_SELF), tc_def_bool(TC_HIDE_SELF));
+	hide_myself = purple_account_get_bool(ma->account, mc_name(TC_HIDE_SELF), mc_def_bool(TC_HIDE_SELF));
 	msg_list = g_list_reverse(msg_list);
 	for(it = g_list_first(msg_list); it; it = g_list_next(it)) {
 
@@ -409,7 +511,7 @@ gint twitter_fetch_new_messages_handler(MbConnData * conn_data, gpointer data)
 			// we still call serv_got_im here, so purple take the message to the log
 			serv_got_im(ma->gc, tlr->name, msg_txt, PURPLE_MESSAGE_RECV, cur_msg->msg_time);
 			// by handling diaplying-im-msg, the message shouldn't be displayed anymore
-			purple_signal_emit(tc_def(TC_PLUGIN), "twitter-message", ma, tlr->name, cur_msg);
+			purple_signal_emit(mc_def(TC_PLUGIN), "twitter-message", ma, tlr->name, cur_msg);
 			g_free(msg_txt);
 		}
 		g_free(id_str);
@@ -434,74 +536,54 @@ gint twitter_fetch_new_messages_handler(MbConnData * conn_data, gpointer data)
 //
 // Check for new message periodically
 //
-void twitter_fetch_new_messages(MbAccount * ta, TwitterTimeLineReq * tlr)
+void twitter_fetch_new_messages(MbAccount * ma, TwitterTimeLineReq * tlr)
 {
 	MbConnData * conn_data;
-	MbHttpData * request;
-	gchar * twitter_host, * user_name;
-	gboolean use_https;
-	gint twitter_port;
 	
 	purple_debug_info(DBGID, "%s called\n", __FUNCTION__);
-
-	twitter_get_user_host(ta, &user_name, &twitter_host);
-	use_https = purple_account_get_bool(ta->account, tc_name(TC_USE_HTTPS), tc_def_bool(TC_USE_HTTPS));
-	if(use_https) {
-		twitter_port = TW_HTTPS_PORT;
-	} else {
-		twitter_port = TW_HTTP_PORT;
-	}
 	
-	conn_data = mb_conn_data_new(ta, twitter_host, twitter_port, twitter_fetch_new_messages_handler, use_https);
-	mb_conn_data_set_retry(conn_data, 0);
-	
-	request = conn_data->request;
-	request->type = HTTP_GET;
-	request->port = twitter_port;
+	conn_data = twitter_init_connection(ma, HTTP_GET, tlr->path, twitter_fetch_new_messages_handler);
 
-	mb_http_data_set_host(request, twitter_host);
-	mb_http_data_set_path(request, tlr->path);
-	mb_http_data_set_fixed_headers(request, twitter_fixed_headers);
-	mb_http_data_set_header(request, "Host", twitter_host);
-	mb_http_data_set_basicauth(request, user_name, purple_account_get_password(ta->account));
 	if(tlr->count > 0) {
 		purple_debug_info(DBGID, "tlr->count = %d\n", tlr->count);
-		mb_http_data_add_param_int(request, "count", tlr->count);
+		mb_http_data_add_param_int(conn_data->request, "count", tlr->count);
 	}
-	if(tlr->use_since_id && (ta->last_msg_id > 0) ) {
-		mb_http_data_add_param_ull(request, "since_id", ta->last_msg_id);
+	if(tlr->use_since_id && (ma->last_msg_id > 0) ) {
+		mb_http_data_add_param_ull(conn_data->request, "since_id", ma->last_msg_id);
+	}
+	if(tlr->screen_name != NULL) {
+		mb_http_data_add_param(conn_data->request, "screen_name", tlr->screen_name);
 	}
 	conn_data->handler_data = tlr;
 	
 	mb_conn_process_request(conn_data);
-	g_free(twitter_host);
-	g_free(user_name);
 }
 
+
+
 //
 // Generate 'fake' buddy list for Twitter
 // For now, we only add TwFriends, TwUsers, and TwPublic
-void twitter_get_buddy_list(TwitterAccount * ta)
+void twitter_get_buddy_list(MbAccount * ma)
 {
 	PurpleBuddy *buddy;
 	TwitterBuddy *tbuddy;
 	PurpleGroup *twitter_group = NULL;
 
-	purple_debug_info(DBGID, "buddy list for account %s\n", ta->account->username);
+	purple_debug_info(DBGID, "buddy list for account %s\n", ma->account->username);
 
 	//Check if the twitter group already exists
-	twitter_group = purple_find_group(tc_def(TC_USER_GROUP));
+	twitter_group = purple_find_group(mc_def(TC_USER_GROUP));
 	
 	// Add timeline as "fake" user
 	// Is TL_FRIENDS already exist?
-	if ( (buddy = purple_find_buddy(ta->account, tc_def(TC_FRIENDS_USER))) == NULL)
-	{
-		purple_debug_info(DBGID, "creating new buddy list for %s\n", tc_def(TC_FRIENDS_USER));
-		buddy = purple_buddy_new(ta->account, tc_def(TC_FRIENDS_USER), NULL);
+	if ( (buddy = purple_find_buddy(ma->account, mc_def(TC_FRIENDS_USER))) == NULL)	{
+		purple_debug_info(DBGID, "creating new buddy list for %s\n", mc_def(TC_FRIENDS_USER));
+		buddy = purple_buddy_new(ma->account, mc_def(TC_FRIENDS_USER), NULL);
 		if (twitter_group == NULL)
 		{
 			purple_debug_info(DBGID, "creating new Twitter group\n");
-			twitter_group = purple_group_new(tc_def(TC_USER_GROUP));
+			twitter_group = purple_group_new(mc_def(TC_USER_GROUP));
 			purple_blist_add_group(twitter_group, NULL);
 		}
 		purple_debug_info(DBGID, "setting protocol-specific buddy information to purplebuddy\n");
@@ -509,43 +591,22 @@ void twitter_get_buddy_list(TwitterAccount * ta)
 			tbuddy = twitter_new_buddy();
 			buddy->proto_data = tbuddy;
 			tbuddy->buddy = buddy;
-			tbuddy->ta = ta;
+			tbuddy->ma = ma;
 			tbuddy->uid = TL_FRIENDS;
-			tbuddy->name = g_strdup(tc_def(TC_FRIENDS_USER));
+			tbuddy->name = g_strdup(mc_def(TC_FRIENDS_USER));
 		}
 		purple_blist_add_buddy(buddy, NULL, twitter_group, NULL);
 	}
-	purple_prpl_got_user_status(ta->account, buddy->name, purple_primitive_get_id_from_type(PURPLE_STATUS_AVAILABLE), NULL);
+	purple_prpl_got_user_status(ma->account, buddy->name, purple_primitive_get_id_from_type(PURPLE_STATUS_AVAILABLE), NULL);
 	// We'll deal with public and users timeline later
 }
 
-gint twitter_verify_authen(MbConnData * conn_data, gpointer data)
-{
-	MbAccount * ma = conn_data->ma;
-	MbHttpData * response = conn_data->response;
-	
-	if(response->status == HTTP_OK) {
-		gint interval = purple_account_get_int(conn_data->ma->account, tc_name(TC_MSG_REFRESH_RATE), tc_def_int(TC_MSG_REFRESH_RATE));
-		
-		purple_connection_set_state(conn_data->ma->gc, PURPLE_CONNECTED);
-		conn_data->ma->state = PURPLE_CONNECTED;
-		twitter_get_buddy_list(conn_data->ma);
-		purple_debug_info(DBGID, "refresh interval = %d\n", interval);
-		conn_data->ma->timeline_timer = purple_timeout_add_seconds(interval, (GSourceFunc)twitter_fetch_all_new_messages, conn_data->ma);
-		twitter_fetch_first_new_messages(conn_data->ma);
-		return 0;
-	} else {
-		// XXX: Crash at the line below
-		purple_connection_set_state(conn_data->ma->gc, PURPLE_DISCONNECTED);
-		conn_data->ma->state = PURPLE_DISCONNECTED;
-		purple_connection_error(ma->gc, _("Authentication error"));
-		return -1;
-	}
-}
-
 MbAccount * mb_account_new(PurpleAccount * acct)
 {
 	MbAccount * ma = NULL;
+	const char * auth_type;
+	int i;
+	const gchar * oauth_token = NULL, * oauth_secret = NULL;
 	
 	purple_debug_info(DBGID, "%s\n", __FUNCTION__);
 	ma = g_new(MbAccount, 1);
@@ -560,6 +621,39 @@ MbAccount * mb_account_new(PurpleAccount * acct)
 	ma->tag = NULL;
 	ma->tag_pos = MB_TAG_NONE;
 	ma->reply_to_status_id = 0;
+	ma->mb_conf = _mb_conf;
+
+	// Cache
+//	ma->cache = mb_cache_new();
+
+	// Auth Type
+	if(mc_name(TC_AUTH_TYPE)) {
+		// Only twitter has this configuration
+		auth_type = purple_account_get_string(acct, mc_name(TC_AUTH_TYPE), mc_def(TC_AUTH_TYPE));
+		if(auth_type) {
+			for(i = 0; i < MB_AUTH_MAX; i++) {
+				if(strcmp(mb_auth_types_str[i], auth_type) == 0) {
+					ma->auth_type = i;
+					break;
+				}
+			}
+		}
+		purple_debug_info(DBGID, "auth_type = %d\n", ma->auth_type);
+	} else {
+		// if there's no configuration, then it means the protocol didn't support it
+		// fall back to http basic authentication
+		ma->auth_type = MB_HTTP_BASICAUTH;
+	}
+
+	// Oauth stuff
+	mb_oauth_init(ma, mc_def(TC_CONSUMER_KEY), mc_def(TC_CONSUMER_SECRET));
+	oauth_token = purple_account_get_string(ma->account, mc_name(TC_OAUTH_TOKEN), NULL);
+	oauth_secret = purple_account_get_string(ma->account, mc_name(TC_OAUTH_SECRET), NULL);
+
+	if(oauth_token && oauth_secret && (strlen(oauth_token) > 0) && (strlen(oauth_secret) > 0)) {
+		mb_oauth_set_token(ma, oauth_token, oauth_secret);
+	}
+
 	acct->gc->proto_data = ma;
 	return ma;
 }
@@ -580,7 +674,7 @@ static void mb_close_connection(gpointer key, gpointer value, gpointer user_data
 static gboolean foreach_remove_expire_idhash(gpointer key, gpointer val, gpointer userdata)
 {
 	MbAccount * ma = (MbAccount *)userdata;
-	unsigned long long msg_id;
+	mb_status_t msg_id;
 
 	msg_id = strtoull(key, NULL, 10);
 	if(ma->last_msg_id >= msg_id) {
@@ -597,7 +691,14 @@ void mb_account_free(MbAccount * ma)
 
 	//purple_debug_info(DBGID, "mb_account_free\n");
 	purple_debug_info(DBGID, "%s\n", __FUNCTION__);
+
+	// Remove cache
+//	mb_cache_free(ma->cache);
+	ma->mb_conf = NULL;
+	ma->cache = NULL;
 	
+	mb_oauth_free(ma);
+
 	if(ma->tag) {
 		g_free(ma->tag);
 		ma->tag = NULL;
@@ -634,59 +735,231 @@ void mb_account_free(MbAccount * ma)
 
 void twitter_login(PurpleAccount *acct)
 {
-	MbAccount *ta = NULL;
-	MbConnData * conn_data = NULL;
-	gchar * twitter_host = NULL;
-	gchar * path = NULL, * user_name = NULL;
-	gboolean use_https = TRUE;
+	MbAccount *ma = NULL;
 	
 	purple_debug_info(DBGID, "twitter_login\n");
 	
 	// Create account data
-	ta = mb_account_new(acct);
+	ma = mb_account_new(acct);
 
 	purple_debug_info(DBGID, "creating id hash for sentid\n");
-	mb_account_get_idhash(acct, TW_ACCT_SENT_MSG_IDS,ta->sent_id_hash);
+	mb_account_get_idhash(acct, TW_ACCT_SENT_MSG_IDS,ma->sent_id_hash);
 
-	purple_debug_info(DBGID, "getting user/host\n");
-	twitter_get_user_host(ta, &user_name, &twitter_host);
-
-	purple_debug_info(DBGID, "user_name = %s\n", user_name);
-	path = g_strdup(purple_account_get_string(ta->account, tc_name(TC_VERIFY_PATH), tc_def(TC_VERIFY_PATH)));
-	use_https = purple_account_get_bool(ta->account, tc_name(TC_USE_HTTPS), tc_def_int(TC_USE_HTTPS));
-	if(use_https) {
-		twitter_port = TW_HTTPS_PORT;
-	} else {
-		twitter_port = TW_HTTP_PORT;
-	}
-	purple_debug_info(DBGID, "path = %s\n", path);
-	
-	conn_data = mb_conn_data_new(ta, twitter_host, twitter_port, twitter_verify_authen, use_https);
-	mb_conn_data_set_retry(conn_data, purple_account_get_int(acct, tc_name(TC_GLOBAL_RETRY), tc_def_int(TC_GLOBAL_RETRY)));
-	
-	conn_data->request->type = HTTP_GET;
-	mb_http_data_set_host(conn_data->request, twitter_host);
-	mb_http_data_set_path(conn_data->request, path);
-	mb_http_data_set_fixed_headers(conn_data->request, twitter_fixed_headers);
-	mb_http_data_set_header(conn_data->request, "Host", twitter_host);
-	mb_http_data_set_basicauth(conn_data->request, user_name, purple_account_get_password(ta->account));
-
-	//purple_proxy_connect(NULL, ta->account, twitter_host, twitter_port, test_connect_cb, NULL);
-	mb_conn_process_request(conn_data);
-	g_free(twitter_host);
-	g_free(user_name);
-	g_free(path);
+	twitter_request_access(ma);
 
 	// connect to twitgin here
 	purple_debug_info(DBGID, "looking for twitgin\n");
 	twitgin_plugin = purple_plugins_find_with_id("gtktwitgin");
 	if(twitgin_plugin) {
 		purple_debug_info(DBGID, "registering twitgin-replying-message signal\n");
-		purple_signal_connect(twitgin_plugin, "twitgin-replying-message", acct, PURPLE_CALLBACK(twitter_on_replying_message), ta);
+		purple_signal_connect(twitgin_plugin, "twitgin-replying-message", acct, PURPLE_CALLBACK(twitter_on_replying_message), ma);
 	}
-	
 }
 
+/**
+ * Request for access token, if needed
+ *
+ * @param ma MbAccount in action
+ */
+void twitter_request_access(MbAccount * ma)
+{
+	const gchar * path = NULL;
+	const gchar * oauth_token = NULL, * oauth_secret = NULL;
+
+	// check if oauth or xauth is already done
+	switch(ma->auth_type) {
+		case MB_OAUTH :
+			// If oauth access request is not done yet, do it now
+			oauth_token = purple_account_get_string(ma->account, mc_name(TC_OAUTH_TOKEN), mc_def(TC_OAUTH_TOKEN));
+			oauth_secret = purple_account_get_string(ma->account, mc_name(TC_OAUTH_SECRET), mc_def(TC_OAUTH_SECRET));
+			if(!oauth_token || !oauth_secret || (strlen(oauth_token) <= 0) || (strlen(oauth_secret) <= 0)) {
+				mb_oauth_init(ma, mc_def(TC_CONSUMER_KEY), mc_def(TC_CONSUMER_SECRET));
+				path = purple_account_get_string(ma->account, mc_name(TC_REQUEST_TOKEN_URL), mc_def(TC_REQUEST_TOKEN_URL));
+				mb_oauth_request_token(ma, path, HTTP_GET, twitter_request_authorize, NULL);
+			} else {
+				twitter_verify_account(ma, NULL);
+			}
+			break;
+		case MB_XAUTH :
+			// support this later
+			break;
+		default :
+			// default is basic auth, which require no request access
+			twitter_verify_account(ma, NULL);
+			break;
+	}
+}
+
+/*
+ * Redirect user to authorization page and wait for user input
+ */
+gint twitter_request_authorize(MbAccount * ma, MbConnData * data, gpointer user_data)
+{
+	const gchar * request_access_path = NULL;
+	gchar * error_msg = NULL;
+	gchar * user_name, * host, *param = NULL, * full_url;
+	gboolean use_https = FALSE;
+
+
+	if( (data->response->status != HTTP_OK) || (!ma->oauth.oauth_token && !ma->oauth.oauth_secret)) {
+		// error
+		if(data->response->content_len > 0) {
+			error_msg = g_strdup(data->response->content->str);
+		} else {
+			error_msg = g_strdup("Unknown error");
+		}
+		mb_conn_error(data, PURPLE_CONNECTION_ERROR_INVALID_SETTINGS, error_msg);
+		g_free(error_msg);
+		return -1;
+	}
+
+	// process the successfully acquire token
+	request_access_path = purple_account_get_string(ma->account, mc_name(TC_AUTHORIZE_URL), mc_def(TC_AUTHORIZE_URL));
+	use_https = purple_account_get_bool(ma->account, mc_name(TC_USE_HTTPS), mc_def_bool(TC_USE_HTTPS));
+	twitter_get_user_host(ma, &user_name, &host);
+	param = g_strdup_printf("oauth_token=%s", ma->oauth.oauth_token);
+	full_url = mb_url_unparse(host, 0, request_access_path, param, use_https);
+
+	g_free(param);
+
+	// Open the web browser and redirect user to the authorization page
+	purple_notify_uri((void *)ma->gc, full_url);
+	g_free(full_url);
+
+	// and wait for user's input for PIN
+	purple_request_input((void *)ma->gc, _("Input your PIN"), _("Please allow " TW_AGENT_SOURCE " to access your account"),
+			_("Please copy the PIN number from the web page"), "", FALSE, FALSE, NULL,
+			_("OK"), G_CALLBACK(twitter_request_authorize_ok_cb),
+			_("Cancel"), NULL, //< We will not handle cancel, user's will have to re-authorize again
+			ma->account, NULL, NULL, ma);
+
+	g_free(user_name);
+	g_free(host);
+
+	// Continue in PIN callback
+	return 0;
+}
+
+void twitter_request_authorize_ok_cb(MbAccount * ma, const char * pin)
+{
+	const gchar * access_token_path = NULL;
+	purple_debug_info(DBGID, "%s called\n", __FUNCTION__);
+	purple_debug_info(DBGID, "got PIN %s\n", pin);
+
+	// Set the PIN
+	mb_oauth_set_pin(ma, pin);
+
+	access_token_path = purple_account_get_string(ma->account, mc_name(TC_ACCESS_TOKEN_URL), mc_def(TC_ACCESS_TOKEN_URL));
+	mb_oauth_request_access(ma, access_token_path, HTTP_POST, twitter_oauth_request_finish, NULL);
+}
+
+/**
+ * Wrapper to call twitter verify account from OAuth realm
+ */
+gint twitter_oauth_request_finish(MbAccount * ma, MbConnData * data, gpointer user_data) {
+	if((data->response->status == HTTP_OK) && ma->oauth.oauth_token && ma->oauth.oauth_secret) {
+		if(ma->oauth.pin) {
+			g_free(ma->oauth.pin);
+			ma->oauth.pin = NULL;
+		}
+		purple_account_set_string(ma->account, mc_name(TC_OAUTH_TOKEN), ma->oauth.oauth_token);
+		purple_account_set_string(ma->account, mc_name(TC_OAUTH_SECRET), ma->oauth.oauth_secret);
+		twitter_verify_account(ma, NULL);
+	} else {
+		if(ma->oauth.oauth_token) g_free(ma->oauth.oauth_token);
+		if(ma->oauth.oauth_secret) g_free(ma->oauth.oauth_secret);
+		ma->oauth.oauth_token = ma->oauth.oauth_secret = NULL;
+		purple_connection_error_reason(ma->gc, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Invalid server response");
+	}
+
+	return 0;
+}
+
+/**
+ * Verify the account with services
+ *
+ * @param ma MbAccount in action
+ */
+void twitter_verify_account(MbAccount * ma, gpointer data)
+{
+	MbConnData * conn_data = NULL;
+	gchar * path = NULL;
+
+	path = g_strdup(purple_account_get_string(ma->account, mc_name(TC_VERIFY_PATH), mc_def(TC_VERIFY_PATH)));
+	purple_debug_info(DBGID, "path = %s\n", path);
+
+	conn_data = twitter_init_connection(ma, HTTP_GET, path, twitter_verify_authen);
+
+	mb_conn_process_request(conn_data);
+	g_free(path);
+}
+
+/**
+ * Call back for account validity verification
+ *
+ * @param conn_data connection data in progress
+ * @param data MbAccount in action
+ */
+gint twitter_verify_authen(MbConnData * conn_data, gpointer data, const char * error)
+{
+	MbAccount * ma = conn_data->ma;
+	MbHttpData * response = conn_data->response;
+
+	if(response->content_len > 0) {
+		purple_debug_info(DBGID, "response = %s\n", response->content->str);
+	}
+	if(response->status == HTTP_OK) {
+		gint interval = purple_account_get_int(conn_data->ma->account, mc_name(TC_MSG_REFRESH_RATE), mc_def_int(TC_MSG_REFRESH_RATE));
+
+		// extract the username and set it
+		if(response->content_len > 0) {
+			xmlnode * top = NULL, * screen_name = NULL;
+			gchar * screen_name_str = NULL;
+			gchar * user_name = NULL, * host = NULL;
+
+			top = xmlnode_from_str(response->content->str, -1);
+			if(top != NULL) {
+				screen_name = xmlnode_get_child(top, "screen_name");
+				if(screen_name) {
+					screen_name_str = xmlnode_get_data_unescaped(screen_name);
+				}
+			}
+			xmlnode_free(top);
+			if(screen_name_str) {
+				purple_debug_info(DBGID, "old username = %s\n", purple_account_get_username(conn_data->ma->account));
+				twitter_get_user_host(conn_data->ma, &user_name, &host);
+				if(host) {
+					// libpurple host is embed in username, so we need to keep it
+					gchar * tmp;
+
+					tmp = g_strdup_printf("%s@%s", screen_name_str, host);
+					purple_account_set_username(conn_data->ma->account, tmp);
+					g_free(tmp);
+				} else {
+					purple_account_set_username(conn_data->ma->account, screen_name_str);
+				}
+				g_free(user_name);
+				g_free(host);
+			} else {
+				purple_debug_info(DBGID, "WARNING! will use username in setting instead\n");
+			}
+			g_free(screen_name_str);
+		}
+
+		// now prepare for timeline refresher
+		purple_connection_set_state(conn_data->ma->gc, PURPLE_CONNECTED);
+		conn_data->ma->state = PURPLE_CONNECTED;
+		twitter_get_buddy_list(conn_data->ma);
+		purple_debug_info(DBGID, "refresh interval = %d\n", interval);
+		conn_data->ma->timeline_timer = purple_timeout_add_seconds(interval, (GSourceFunc)twitter_fetch_all_new_messages, conn_data->ma);
+		twitter_fetch_first_new_messages(conn_data->ma);
+		return 0;
+	} else {
+		// XXX: Crash at the line below
+		mb_conn_error(conn_data, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Authentication error");
+		return -1;
+	}
+}
 
 void twitter_close(PurpleConnection *gc)
 {
@@ -708,22 +981,40 @@ void twitter_close(PurpleConnection *gc)
 	gc->proto_data = NULL;
 }
 
-gint twitter_send_im_handler(MbConnData * conn_data, gpointer data)
+gint twitter_send_im_handler(MbConnData * conn_data, gpointer data, const char * error)
 {
 	MbAccount * ma = conn_data->ma;
 	MbHttpData * response = conn_data->response;
-	gchar * id_str = NULL;
+	gchar * id_str = NULL, * who = (gchar *)data;
 	xmlnode * top, *id_node;
 	
-	purple_debug_info(DBGID, "send_im_handler\n");
-	
-	if(response->status != 200) {
-		purple_debug_info(DBGID, "http error\n");
-		//purple_debug_info(DBGID, "http data = #%s#\n", response->content->str);
+	purple_debug_info(DBGID, "%s\n", __FUNCTION__);
+
+	if(error) {
+		// don't need to handle network error
+		if(mb_conn_max_retry_reach(conn_data)) {
+			g_free(who);
+		}
 		return -1;
 	}
 	
-	if(!purple_account_get_bool(ma->account, tc_name(TC_HIDE_SELF), tc_def_bool(TC_HIDE_SELF))) {
+	if(response->status != HTTP_OK) {
+		purple_debug_info(DBGID, "http error\n");
+		if(response->content_len > 0) {
+			purple_debug_info(DBGID, "response = %s\n", response->content->str);
+		}
+		//purple_debug_info(DBGID, "http data = #%s#\n", response->content->str);
+		if(mb_conn_max_retry_reach(conn_data)) {
+			serv_got_im(ma->gc, who, _("error sending status"), PURPLE_MESSAGE_SYSTEM, time(NULL));
+			g_free(who);
+		}
+		return -1;
+	}
+	
+	// We don't need who from this point on
+	g_free(who);
+
+	if(!purple_account_get_bool(ma->account, mc_name(TC_HIDE_SELF), mc_def_bool(TC_HIDE_SELF))) {
 		return 0;
 	}
 	
@@ -760,24 +1051,23 @@ gint twitter_send_im_handler(MbConnData * conn_data, gpointer data)
 
 int twitter_send_im(PurpleConnection *gc, const gchar *who, const gchar *message, PurpleMessageFlags flags)
 {
-	TwitterAccount * ta = gc->proto_data;
+	MbAccount * ma = gc->proto_data;
 	MbConnData * conn_data = NULL;
-	gchar * post_data = NULL, * tmp_msg_txt = NULL, * user_name = NULL;
-	gint msg_len, twitter_port, len;
-	gchar * twitter_host, * path;
-	gboolean use_https;
+	gchar * tmp_msg_txt = NULL;
+	gint msg_len;
+	gchar * path;
 	
-	purple_debug_info(DBGID, "send_im\n");
+	purple_debug_info(DBGID, "%s called, who = %s, message = %s, flag = %d\n", __FUNCTION__, who, message, flags);
 
 	// prepare message to send
-	tmp_msg_txt = g_strdup(purple_url_encode(g_strchomp(purple_markup_strip_html(message))));
-	if(ta->tag) {
+	tmp_msg_txt = g_strdup(g_strchomp(purple_markup_strip_html(message)));
+	if(ma->tag) {
 		gchar * new_msg_txt;
 
-		if(ta->tag_pos == MB_TAG_PREFIX) {
-			new_msg_txt  = g_strdup_printf("%s %s", ta->tag, tmp_msg_txt);
+		if(ma->tag_pos == MB_TAG_PREFIX) {
+			new_msg_txt  = g_strdup_printf("%s %s", ma->tag, tmp_msg_txt);
 		} else {
-			new_msg_txt  = g_strdup_printf("%s %s", tmp_msg_txt, ta->tag);
+			new_msg_txt  = g_strdup_printf("%s %s", tmp_msg_txt, ma->tag);
 		}
 		g_free(tmp_msg_txt);
 		tmp_msg_txt = new_msg_txt;
@@ -787,26 +1077,11 @@ int twitter_send_im(PurpleConnection *gc, const gchar *who, const gchar *message
 	purple_debug_info(DBGID, "sending message %s\n", tmp_msg_txt);
 	
 	// connection
-	twitter_get_user_host(ta, &user_name, &twitter_host);
-	path = g_strdup(purple_account_get_string(ta->account, tc_name(TC_STATUS_UPDATE), tc_def(TC_STATUS_UPDATE)));
-	use_https = purple_account_get_bool(ta->account, tc_name(TC_USE_HTTPS), tc_def_bool(TC_USE_HTTPS));
-	
-	if(use_https) {
-		twitter_port = TW_HTTPS_PORT;
-	} else {
-		twitter_port = TW_HTTP_PORT;
-	}
-	conn_data = mb_conn_data_new(ta, twitter_host, twitter_port, twitter_send_im_handler, use_https);
-	mb_conn_data_set_retry(conn_data, 0);
-	conn_data->request->type = HTTP_POST;
-	mb_http_data_set_host(conn_data->request, twitter_host);
-	mb_http_data_set_path(conn_data->request, path);
-	mb_http_data_set_fixed_headers(conn_data->request, twitter_fixed_headers);
-	mb_http_data_set_header(conn_data->request, "Content-Type", "application/x-www-form-urlencoded");
-	mb_http_data_set_header(conn_data->request, "Host", twitter_host);
-	mb_http_data_set_basicauth(conn_data->request, 	user_name,purple_account_get_password(ta->account));
+	path = g_strdup(purple_account_get_string(ma->account, mc_name(TC_STATUS_UPDATE), mc_def(TC_STATUS_UPDATE)));
+	conn_data = twitter_init_connection(ma, HTTP_POST, path, twitter_send_im_handler);
+	conn_data->handler_data = g_strdup(who);
 
-	if(ta->reply_to_status_id > 0) {
+	if(ma->reply_to_status_id > 0) {
 		int i;
 		gboolean do_reply = FALSE;
 		// do not add reply tag if the message does not contains @ in the front
@@ -821,23 +1096,29 @@ int twitter_send_im(PurpleConnection *gc, const gchar *who, const gchar *message
 			}
 		}
 		if(do_reply) {
-			purple_debug_info(DBGID, "setting in_reply_to_status_id = %llu\n", ta->reply_to_status_id);
-			mb_http_data_add_param_ull(conn_data->request, "in_reply_to_status_id", ta->reply_to_status_id);
-			ta->reply_to_status_id = 0;
+			purple_debug_info(DBGID, "setting in_reply_to_status_id = %llu\n", ma->reply_to_status_id);
+			mb_http_data_add_param_ull(conn_data->request, "in_reply_to_status_id", ma->reply_to_status_id);
+			ma->reply_to_status_id = 0;
 		} else {
-			ta->reply_to_status_id = 0;
+			ma->reply_to_status_id = 0;
 		}
 	}
 	
+//	mb_http_data_set_header(conn_data->request, "Content-Type", "application/x-www-form-urlencoded");
+	mb_http_data_set_content_type(conn_data->request, "application/x-www-form-urlencoded");
+
+	mb_http_data_add_param(conn_data->request, "status", tmp_msg_txt);
+	mb_http_data_add_param(conn_data->request, "source", TW_AGENT_SOURCE);
+
+	/*
 	post_data = g_malloc(TW_MAXBUFF);
 	len = snprintf(post_data, TW_MAXBUFF, "status=%s&source=" TW_AGENT_SOURCE, tmp_msg_txt);
 	mb_http_data_set_content(conn_data->request, post_data, len);
+	*/
+	//	g_free(post_data);
 	
 	mb_conn_process_request(conn_data);
-	g_free(twitter_host);
-	g_free(user_name);
 	g_free(path);
-	g_free(post_data);
 	g_free(tmp_msg_txt);
 	
 	return msg_len;
@@ -863,7 +1144,7 @@ void twitter_set_status(PurpleAccount *acct, PurpleStatus *status)
 
 }
 
-void * twitter_on_replying_message(gchar * proto, unsigned long long msg_id, MbAccount * ma)
+void * twitter_on_replying_message(gchar * proto, mb_status_t msg_id, MbAccount * ma)
 {
 	purple_debug_info(DBGID, "%s called!\n", __FUNCTION__);
 	purple_debug_info(DBGID, "setting reply_to_id (was %llu) to %llu\n", ma->reply_to_status_id, msg_id);
@@ -874,45 +1155,36 @@ void * twitter_on_replying_message(gchar * proto, unsigned long long msg_id, MbA
 /*
 *  Favourite Handler
 */
-void twitter_favorite_message(MbAccount * ta, gchar * msg_id)
+void twitter_favorite_message(MbAccount * ma, gchar * msg_id)
 {
 
 	// create new connection and call API POST
-        MbConnData * conn_data;
-        MbHttpData * request;
-        gchar * twitter_host, * user_name, * path;
-        gboolean use_https;
-        gint twitter_port;
+	MbConnData * conn_data;
+	gchar * path;
 
-	user_name = g_strdup_printf("%s", purple_account_get_username(ta->account));
-	twitter_host = g_strdup_printf("%s", "twitter.com");
 	path = g_strdup_printf("/favorites/create/%s.xml", msg_id);
 
-        use_https = TRUE; 
-        if(use_https) {
-                twitter_port = TW_HTTPS_PORT;
-        } else {
-                twitter_port = TW_HTTP_PORT;
-        }
+	conn_data = twitter_init_connection(ma, HTTP_POST, path, NULL);
 
-        conn_data = mb_conn_data_new(ta, twitter_host, twitter_port, NULL, use_https);
-        mb_conn_data_set_retry(conn_data, 0);
+	mb_conn_process_request(conn_data);
+	g_free(path);
 
-        request = conn_data->request;
-        request->type = HTTP_POST;
-        request->port = twitter_port;
+}
 
-	mb_http_data_set_host(request, twitter_host);
-        mb_http_data_set_path(request, path);
-        mb_http_data_set_fixed_headers(request, twitter_fixed_headers);
-        mb_http_data_set_header(request, "Host", twitter_host);
-        mb_http_data_set_basicauth(request, user_name, purple_account_get_password(ta->account));
+/*
+*  Retweet API Handler
+*/
+void twitter_retweet_message(MbAccount * ma, gchar * msg_id)
+{
 
-        //conn_data->handler_data = tlr;
+	// create new connection and call API POST
+	MbConnData * conn_data;
+	gchar * path;
 
-        mb_conn_process_request(conn_data);
-        g_free(twitter_host);
-        g_free(user_name);
+	path = g_strdup_printf("/statuses/retweet/%s.xml", msg_id);
+
+	conn_data = twitter_init_connection(ma, HTTP_POST, path, NULL);
+	mb_conn_process_request(conn_data);
 	g_free(path);
 
 }
