@@ -1,6 +1,6 @@
 /*
  * Twitgin - A GUI support of libtwitter/microblog-purple for Conversation dialog
- * Copyright (C) 2008 Chanwit Kaewkasi <chanwit@gmail.com>
+ * Copyright (C) 2008-2010 Chanwit Kaewkasi  <chanwit@gmail.com>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -23,6 +23,7 @@
 
 #include <string.h>
 #include <stdlib.h>
+#include <ctype.h>
 #include <time.h>
 #include <glib.h>
 #include <glib/gi18n.h>
@@ -60,10 +61,10 @@
 #define DBGID "twitgin"
 #define PURPLE_MESSAGE_TWITGIN 0x1000
 
-// Dummy tw_conf to resolve external symbol link
-TwitterConfig * _tw_conf = NULL;
-
 static PurplePlugin * twitgin_plugin = NULL;
+MbConfig * _mb_conf = NULL;
+
+gchar * twitter_reformat_msg(MbAccount * ta, const TwitterMsg * msg, PurpleConversation * conv);
 
 static void twitgin_entry_buffer_on_changed(PidginConversation *gtkconv) {
 	GtkTextIter start;
@@ -147,16 +148,18 @@ static gboolean twittgin_uri_handler(const char *proto, const char *cmd_arg, GHa
 	const char * cmd = cmd_arg;
 	char *acct_id = g_hash_table_lookup(params, "account");	
 	PurpleAccount *acct = NULL;
+	MbAccount * ma;
 	PurpleConversation * conv = NULL;
 	PidginConversation * gtkconv;
 	int proto_id = 0;
 	gchar * src = NULL;
-	const char * decoded_rt;
-	char * unescaped_rt;
+	//const char * decoded_rt;
+	//char * unescaped_rt;
 
 	purple_debug_info(DBGID, "twittgin_uri_handler\n");	
 
 	// do not need to test, because the conversation window must be open before one can click
+	// XXX: We need better function to search for the account!
 	if (g_ascii_strcasecmp(proto, "tw") == 0) {
 		proto_id = TWITTER_PROTO;
 		acct = purple_accounts_find(acct_id, "prpl-mbpurple-twitter"); 
@@ -169,7 +172,7 @@ static gboolean twittgin_uri_handler(const char *proto, const char *cmd_arg, GHa
 		purple_debug_info(DBGID, "no src specified\n");
 		switch(proto_id) {
 			case TWITTER_PROTO :
-				src = "twitter.com";
+				src = "api.twitter.com";
 				break;
 			case IDENTICA_PROTO :
 				src = "identi.ca";
@@ -184,12 +187,13 @@ static gboolean twittgin_uri_handler(const char *proto, const char *cmd_arg, GHa
 
 	if ( acct && (proto_id > 0) ) {
 		purple_debug_info(DBGID, "found account with libtwitter, proto_id = %d\n", proto_id);
+		ma = (MbAccount *)acct->gc->proto_data;
 
 		/* tw:rep?to=sender */
 		if (!g_ascii_strcasecmp(cmd, "reply")) {
 			gchar * sender, *tmp;
 			gchar * name_to_reply;
-			unsigned long long msg_id = 0;
+			mb_status_t msg_id = 0;
 
 			conv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_ANY, src, acct);
 			purple_debug_info(DBGID, "conv = %p\n", conv);
@@ -212,6 +216,8 @@ static gboolean twittgin_uri_handler(const char *proto, const char *cmd_arg, GHa
 
 		// retweet hack !
 		if (!g_ascii_strcasecmp(cmd, "rt")) {
+			/* We obsolete this function to use twitter retweet API instead (06-May-2010)
+			
 			gchar * message, * from, * retweet_message;
 
 			conv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_ANY, src, acct);
@@ -227,17 +233,24 @@ static gboolean twittgin_uri_handler(const char *proto, const char *cmd_arg, GHa
 			g_free(unescaped_rt);
 			g_free(retweet_message);
 			return TRUE;
-		}
-
-		// favorite hack !
-		if (!g_ascii_strcasecmp(cmd, "fav")) {
-			MbAccount * ta;
+			*/ 
 			gchar * msg_id;
 
 			conv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_ANY, src, acct);
 			msg_id = g_hash_table_lookup(params, "id");
-			ta = mb_account_new(acct);
-			twitter_favorite_message(ta, msg_id);
+			twitter_retweet_message(ma, msg_id);
+			purple_conv_im_write(PURPLE_CONV_IM(conv), NULL, g_strdup_printf("message %s is retweeted", msg_id), PURPLE_MESSAGE_SYSTEM, time(NULL));
+
+			return TRUE;
+		}
+
+		// favorite hack !
+		if (!g_ascii_strcasecmp(cmd, "fav")) {
+			gchar * msg_id;
+
+			conv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_ANY, src, acct);
+			msg_id = g_hash_table_lookup(params, "id");
+			twitter_favorite_message(ma, msg_id);
 			purple_conv_im_write(PURPLE_CONV_IM(conv), NULL, g_strdup_printf("message %s is favorited", msg_id), PURPLE_MESSAGE_SYSTEM, time(NULL));
 
 			return TRUE;
@@ -267,11 +280,12 @@ static void * twitgin_notify_uri(const char *uri) {
  * Modify sending message in the same way as receiving message
  * This should be done only for message generated by self
  */
-gboolean twitgin_on_displaying(PurpleAccount * account, const char * who, char ** msg, PurpleConversation * conv, PurpleMessageFlags flags)
+gboolean twitgin_on_tweet_send(PurpleAccount * account, const char * who, char ** msg, PurpleConversation * conv, PurpleMessageFlags flags)
 {
 	MbAccount * ma = account->gc->proto_data;
 	char * retval;
 	TwitterMsg twitter_msg;
+	gchar * username = NULL;
 
 	// Do not edit msg from these
 	if ((!is_twitter_conversation(conv)) || (flags & PURPLE_MESSAGE_SYSTEM) ) {
@@ -282,20 +296,34 @@ gboolean twitgin_on_displaying(PurpleAccount * account, const char * who, char *
 		if (flags & PURPLE_MESSAGE_SEND) {
 			purple_debug_info(DBGID, "data being displayed = %s, from = %s, flags = %x\n", (*msg), who, flags);
 			purple_debug_info(DBGID, "conv account = %s, name = %s, title = %s\n", purple_account_get_username(conv->account), conv->name, conv->title);
-			purple_debug_info(DBGID, "data not from myself\n");
+			purple_debug_info(DBGID, "sending text IM\n");
+
 			twitter_msg.id = 0;
 			twitter_msg.avatar_url = NULL;
-			twitter_msg.from = NULL; //< force the plug-in not displaying own name
+//			twitter_msg.from = NULL; //< force the plug-in not displaying own name
+			twitter_get_user_host(ma, &username, NULL);
+			twitter_msg.from = username;
 			twitter_msg.msg_txt = (*msg);
-			twitter_msg.msg_time = 0;
+			twitter_msg.msg_time = time(NULL);
 			twitter_msg.flag = 0;
 			twitter_msg.flag |= TW_MSGFLAG_DOTAG;
+
 			purple_debug_info(DBGID, "going to modify message\n");
-			retval = twitter_reformat_msg(ma, &twitter_msg, NULL, FALSE); //< do not reply to myself
+			retval = twitter_reformat_msg(ma, &twitter_msg, conv); //< do not reply to myself
 			purple_debug_info(DBGID, "new data = %s\n", retval);
-			g_free(*msg);
-			(*msg) = retval;
-			return FALSE;
+
+			purple_conv_im_write(PURPLE_CONV_IM(conv),
+					twitter_msg.from, retval,
+					PURPLE_MESSAGE_RECV | PURPLE_MESSAGE_TWITGIN | PURPLE_MESSAGE_RAW | PURPLE_MESSAGE_NO_LOG | PURPLE_MESSAGE_NICK,
+					twitter_msg.msg_time);
+			g_free(username);
+
+			// Discard the message, let the reformat message handle it
+			return TRUE;
+			//			g_free(*msg);
+			//			(*msg) = retval;
+			//			return FALSE;
+
 			/*
 		} else if( (flags & PURPLE_MESSAGE_SYSTEM) || (flags & PURPLE_MESSAGE_NO_LOG) ||
 				(flags & PURPLE_MESSAGE_ERROR))
@@ -350,24 +378,17 @@ gchar * format_datetime(PurpleConversation * conv, time_t mtime) {
         }
 
 	return mdate;
-	
 }
 
 /*
  * Hack the message display, redirect from normal process (on displaying event) and push them back
  */
-void twitgin_on_tweet(MbAccount * ta, gchar * name, TwitterMsg * cur_msg) {
+void twitgin_on_tweet_recv(MbAccount * ta, gchar * name, TwitterMsg * cur_msg) {
 
 	PurpleConversation * conv;
-	gboolean reply_link = purple_prefs_get_bool(TW_PREF_REPLY_LINK);
 	gchar * fmt_txt = NULL, * tmp = NULL;
-	gchar * displaying_txt = NULL;
-	gchar * linkify_txt = NULL;
-	gchar * fav_txt = NULL, * rt_txt = NULL, * datetime_txt = NULL;
-	const char * embed_rt_txt = NULL;
-	const gchar * account = (const gchar *)purple_account_get_username(ta->account);
-	const char * uri_txt = mb_get_uri_txt(ta->account);
 
+	// Create new conversation if none exist
 	conv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_ANY, name, ta->account);
 	if (conv == NULL) {
 		conv = purple_conversation_new(PURPLE_CONV_TYPE_IM, ta->account, name);
@@ -379,63 +400,255 @@ void twitgin_on_tweet(MbAccount * ta, gchar * name, TwitterMsg * cur_msg) {
 	g_free(cur_msg->msg_txt);
 	cur_msg->msg_txt = tmp;
 
-	fmt_txt = twitter_reformat_msg(ta, cur_msg, name, reply_link);
+	fmt_txt = twitter_reformat_msg(ta, cur_msg, conv);
 	purple_debug_info(DBGID, "fmted text msg = ##%s##\n", fmt_txt);
 
+	purple_conv_im_write(PURPLE_CONV_IM(conv),
+			cur_msg->from, fmt_txt,
+			PURPLE_MESSAGE_RECV | PURPLE_MESSAGE_TWITGIN | PURPLE_MESSAGE_RAW | PURPLE_MESSAGE_NO_LOG | PURPLE_MESSAGE_NICK,
+			cur_msg->msg_time);
+
+	g_free(fmt_txt);
+}
+
+static void twitter_update_link(MbAccount * ta, GString * msg, char sym, const char * name)
+{
+	char * user_name;
+	gboolean user_name_eq_name = FALSE;
+
+	twitter_get_user_host(ta, &user_name, NULL);
+	purple_debug_info(DBGID, "symbol = %c, name = %s, user_name = %s\n", sym, name, user_name);
+	if(strcmp(name, user_name) == 0) {
+		purple_debug_info(DBGID, "name and username is equal\n");
+		user_name_eq_name = TRUE;
+	}
+	if(user_name_eq_name) g_string_append_printf(msg, "<i><b>");
+	if(strcmp(ta->account->protocol_id, "prpl-mbpurple-twitter") == 0) {
+		if(sym == '@') {
+			g_string_append_printf(msg, "@<a href=\"http://twitter.com/%s\">%s</a>",name,name);
+		} else if(sym == '#') {
+			g_string_append_printf(msg, "#<a href=\"http://search.twitter.com/search?q=%%23%s\">%s</a>",name,name);
+		}
+	} else if(strcmp(ta->account->protocol_id, "prpl-mbpurple-identica") == 0) {
+		if(sym == '@') {
+			g_string_append_printf(msg, "@<a href=\"http://identi.ca/%s\">%s</a>",name,name);
+		} else if(sym == '#') {
+			g_string_append_printf(msg, "#<a href=\"http://identi.ca/tag/%s\">%s</a>",name,name);
+		}
+	} else {
+		g_string_append_printf(msg, "%c%s", sym, name);
+	}
+	if(user_name_eq_name) g_string_append_printf(msg, "</b></i>");
+	g_free(user_name);
+}
+
+/**
+ * Build a link to status ID base on protocol number
+ *
+ * @param ma MbAccount in action
+ * @param msg message
+ * @param data generic data
+ */
+gchar * twitter_build_status_link(MbAccount * ma, const TwitterMsg * msg, gpointer data) {
+	if(strcmp(ma->account->protocol_id, "prpl-mbpurple-twitter") == 0) {
+		return g_strdup_printf("http://twitter.com/%s/status/%llu", msg->from, (unsigned long long)msg->id);
+	}
+	return NULL;
+}
+
+/*
+ * Reformat text message and makes it looks nicer
+ *
+ * @retval newly allocated buffer of string, need to be freed after used
+ */
+char * twitter_reformat_msg(MbAccount * ma, const TwitterMsg * msg, PurpleConversation * conv)
+{
+	gchar * username = NULL;
+	const gchar * uri_txt = mb_get_uri_txt(ma->account);
+	gchar * fmt_txt = NULL;
+	gchar * linkify_txt = NULL;
+	gchar * fav_txt = NULL;
+	gchar * rt_txt = NULL;
+	gchar * datetime_txt = NULL;
+	gchar * displaying_txt = NULL;
+	GString * output;
+	gchar * src = NULL;
+	gchar * name = NULL, *name_color = NULL;
+	gchar sym, old_char, previous_char;
+	int i = 0, j = 0;
+	gboolean from_eq_username = FALSE;
+	//const char * embed_rt_txt = NULL;
+	gboolean reply_link = purple_prefs_get_bool(TW_PREF_REPLY_LINK);
+	const gchar * account = (const gchar *)purple_account_get_username(ma->account);
+
+	purple_debug_info(DBGID, "%s\n", __FUNCTION__);
+
+	twitter_get_user_host(ma, &username, NULL);
+	output = g_string_new("");
+
+	// tag for the first thing
+	purple_debug_info(DBGID, "checking for tag\n");
+	if( (msg->flag & TW_MSGFLAG_DOTAG) && ma->tag ) {
+		purple_debug_info(DBGID, "do the tagging of message, for the tag %s\n", ma->tag);
+		if(ma->tag_pos == MB_TAG_PREFIX) {
+			src = g_strdup_printf("%s %s", ma->tag, msg->msg_txt);
+		} else {
+			src = g_strdup_printf("%s %s", msg->msg_txt, ma->tag);
+		}
+	} else {
+		purple_debug_info(DBGID, "not doing the tagging of message\n");
+		src = g_strdup(msg->msg_txt);
+	}
+
+	// color of name
+	purple_debug_info(DBGID, "changing colours\n");
+	if(msg->from) {
+		if( strcmp(msg->from, username) == 0) {
+			from_eq_username = TRUE;
+			purple_debug_info(DBGID, "self generated message, %s, %s\n", msg->from, username);
+		}
+	}
+	// switch colour for ourself
+	if(from_eq_username) {
+		name_color = g_strdup("darkred");
+	} else {
+		name_color = g_strdup("darkblue");
+	}
+	g_string_append_printf(output, "<font color=\"%s\"><b>", name_color);
+	//	self-filter is not possible now
+	//	if(strcmp(msg->from, username) != 0) {
+	uri_txt = mb_get_uri_txt(ma->account);
+	if(reply_link && conv && uri_txt) {
+		if(from_eq_username) {
+			g_string_append_printf(output, "<i>");
+		}
+		/*
+			purple_debug_info(DBGID, "current output = %s\n", output->str);
+			purple_debug_info(DBGID, "url text = %s\n", mb_get_uri_txt(ta->account));
+			purple_debug_info(DBGID, "conversation name = %s\n", conv_name);
+			purple_debug_info(DBGID, "from = %s\n", msg->from);
+			purple_debug_info(DBGID, "username = %s\n", username);
+			purple_debug_info(DBGID, "id = %llu\n", msg->id);
+		 */
+		if(msg->id > 0) {
+#if PURPLE_VERSION_CHECK(2, 6, 0)
+			g_string_append_printf(output, "<a href=\"%s:///reply?src=%s&to=%s&account=%s&id=%llu\">%s</a>:", uri_txt, conv->name, msg->from, account, msg->id, msg->from);
+#else
+			g_string_append_printf(output, "<a href=\"%s:reply?src=%s&to=%s&account=%s&id=%llu\">%s</a>:", uri_txt, conv->name, msg->from, account, msg->id, msg->from);
+#endif
+		} else {
+			g_string_append_printf(output, "%s:", msg->from);
+		}
+		if(from_eq_username) {
+			g_string_append_printf(output, "</i>");
+		}
+	} else {
+		g_string_append_printf(output, "%s:", msg->from);
+	}
+	//	}
+	g_string_append_printf(output, "</b></font> ");
+	g_free(name_color);
+
+
+	purple_debug_info(DBGID, "display msg = %s\n", output->str);
+	purple_debug_info(DBGID, "source msg = %s\n", src);
+
+	// now search message text and look for things to highlight
+	previous_char = src[i];
+	while(src[i] != '\0') {
+		//purple_debug_info(DBGID, "previous_char = %c, src[i] == %c\n", previous_char, src[i]);
+		if( (i == 0 || isspace(previous_char)) &&
+			((src[i] == '@') || (src[i] == '#')) )
+		{
+			//purple_debug_info(DBGID, "reformat_msg: sym = %c\n", src[i]);
+			sym = src[i];
+			// if it's a proper name, extract it
+			i++;
+			j = i;
+			while((src[j] != '\0') && (!isspace(src[j]) && !strchr("!@#$%^&*()-=+[]{};:'\"<>?,./`~", src[j]))) {
+				j++;
+			}
+			if(i == j) {
+				// empty string
+				g_string_append_c(output, sym);
+				continue;
+			}
+			old_char = src[j];
+			src[j] = '\0';
+			name = &src[i];
+			twitter_update_link(ma, output, sym, name);
+			src[j] = old_char;
+			i = j;
+			previous_char = src[i-1];
+		} else {
+			g_string_append_c(output, src[i]);
+			previous_char = src[i];
+			i++;
+		}
+	}
+
+	// end of text formatting
+	g_free(username);
+	g_free(src);
+	fmt_txt = g_string_free(output, FALSE);
+
+	//
+	// Now formatting all links
+	//
+
 	// need to manually linkify text since we are going to send RAW message
 	linkify_txt = purple_markup_linkify(fmt_txt);
 
 	if(uri_txt) {
 		// display favorite link, if enabled
-		if(purple_prefs_get_bool(TW_PREF_FAV_LINK)) {
+		if( (msg->id > 0) && purple_prefs_get_bool(TW_PREF_FAV_LINK)) {
 #if PURPLE_VERSION_CHECK(2, 6, 0)
-			fav_txt = g_strdup_printf(" <a href=\"%s:///fav?src=%s&account=%s&id=%llu\">*</a>", uri_txt, name, account, cur_msg->id);
+			fav_txt = g_strdup_printf(" <a href=\"%s:///fav?src=%s&account=%s&id=%llu\">*</a>", uri_txt, conv->name, account, msg->id);
 #else
-			fav_txt = g_strdup_printf(" <a href=\"%s:fav?src=%s&account=%s&id=%llu\">*</a>", uri_txt, name, account, cur_msg->id);
+			fav_txt = g_strdup_printf(" <a href=\"%s:fav?src=%s&account=%s&id=%llu\">*</a>", uri_txt, conv->name, account, msg->id);
 #endif
 		}
 
 		// display rt link, if enabled
-		if(purple_prefs_get_bool(TW_PREF_RT_LINK)) {
+		if( (msg->id > 0) && purple_prefs_get_bool(TW_PREF_RT_LINK) && !msg->is_protected) {
 			// text for retweet url
-			embed_rt_txt = purple_url_encode(cur_msg->msg_txt);
-			purple_debug_info(DBGID, "url embed text for retweet = ##%s##\n", embed_rt_txt);
+			//embed_rt_txt = purple_url_encode(msg->msg_txt);
+			//purple_debug_info(DBGID, "url embed text for retweet = ##%s##\n", embed_rt_txt);
 
 #if PURPLE_VERSION_CHECK(2, 6, 0)
-			rt_txt = g_strdup_printf(" <a href=\"%s:///rt?src=%s&account=%s&from=%s&msg=%s\">rt<a>", uri_txt, name, account, cur_msg->from, embed_rt_txt);
+			//rt_txt = g_strdup_printf(" <a href=\"%s:///rt?src=%s&account=%s&from=%s&msg=%s\">rt<a>", uri_txt, conv->name, account, msg->from, embed_rt_txt);
+			rt_txt = g_strdup_printf(" <a href=\"%s:///rt?src=%s&account=%s&id=%llu\">rt<a>", uri_txt, conv->name, account, msg->id);
 #else
-			rt_txt = g_strdup_printf(" <a href=\"%s:rt?src=%s&account=%s&from=%s&msg=%s\">rt<a>", uri_txt, name, account, cur_msg->from, embed_rt_txt);
+			//rt_txt = g_strdup_printf(" <a href=\"%s:rt?src=%s&account=%s&from=%s&msg=%s\">rt<a>", uri_txt, conv->name, account, msg->from, embed_rt_txt);
+			rt_txt = g_strdup_printf(" <a href=\"%s:rt?src=%s&account=%s&id=%llu\">rt<a>", uri_txt, conv->name, account, msg->id);
 #endif
 		}
+	}
+	if(conv && (msg->msg_time > 0)) {
+		gchar * url = twitter_build_status_link(ma, msg, NULL);
 
 		//display link to message status in the timestamp
-		if(purple_prefs_get_bool(TW_PREF_MS_LINK)) {
-			
-			datetime_txt = g_strdup_printf("<FONT COLOR=\"#cc0000\"><a href=\"http://twitter.com/%s/status/%llu\">%s</a></FONT>", cur_msg->from, cur_msg->id, format_datetime(conv, cur_msg->msg_time)); 
+		if((msg->id > 0) && purple_prefs_get_bool(TW_PREF_MS_LINK) && url) {
+			datetime_txt = g_strdup_printf("<FONT COLOR=\"#cc0000\"><a href=\"http://twitter.com/%s/status/%llu\">%s</a></FONT> ", msg->from, msg->id, format_datetime(conv, msg->msg_time));
+		} else {
+			datetime_txt = g_strdup_printf("<FONT COLOR=\"#cc0000\">%s</FONT> ", format_datetime(conv, msg->msg_time));
 		}
-		else {
-			datetime_txt = g_strdup_printf("<FONT COLOR=\"#cc0000\">%s</FONT>", format_datetime(conv, cur_msg->msg_time)); 
-		}
-		
-		displaying_txt = g_strdup_printf("%s %s%s%s", datetime_txt, linkify_txt, 
-			fav_txt ? fav_txt : "", rt_txt ? rt_txt : "");
-
-		if(fav_txt) g_free(fav_txt);
-		if(rt_txt) g_free(rt_txt);
-		if(datetime_txt) g_free(datetime_txt);
-	} else {
-		displaying_txt = g_strdup_printf("%s %s", format_datetime(conv, cur_msg->msg_time), linkify_txt);
-
+		if(url) g_free(url);
 	}
-	purple_debug_info(DBGID, "displaying text = ##%s##\n", displaying_txt);
-	purple_conv_im_write(PURPLE_CONV_IM(conv),
-		cur_msg->from, displaying_txt, 
-		PURPLE_MESSAGE_RECV | PURPLE_MESSAGE_TWITGIN | PURPLE_MESSAGE_RAW | PURPLE_MESSAGE_NO_LOG | PURPLE_MESSAGE_NICK,
-		cur_msg->msg_time);
 
-	g_free(displaying_txt);
+	displaying_txt = g_strdup_printf("%s%s%s%s", datetime_txt ? datetime_txt : "",
+			linkify_txt, fav_txt ? fav_txt : "", rt_txt ? rt_txt : "");
+
+	if(fav_txt) g_free(fav_txt);
+	if(rt_txt) g_free(rt_txt);
+	if(datetime_txt) g_free(datetime_txt);
+
+	purple_debug_info(DBGID, "displaying text = ##%s##\n", displaying_txt);
+
 	g_free(linkify_txt);
 	g_free(fmt_txt);
+
+	return displaying_txt;
 }
 
 #if PURPLE_VERSION_CHECK(2, 6, 0)
@@ -494,7 +707,7 @@ static gboolean plugin_load(PurplePlugin *plugin)
 #endif
 	purple_signal_connect(purple_get_core(), "uri-handler", plugin, PURPLE_CALLBACK(twittgin_uri_handler), NULL);
 
-	purple_signal_connect(pidgin_conversations_get_handle(), "displaying-im-msg", plugin, PURPLE_CALLBACK(twitgin_on_displaying), NULL);
+	purple_signal_connect(pidgin_conversations_get_handle(), "displaying-im-msg", plugin, PURPLE_CALLBACK(twitgin_on_tweet_send), NULL);
 
 	// twitter
 	// handle all mbpurple plug-in
@@ -503,7 +716,7 @@ static gboolean plugin_load(PurplePlugin *plugin)
 		prpl_plugin = plugins->data;
 		if( (prpl_plugin->info->id != NULL) && (strncmp(prpl_plugin->info->id, "prpl-mbpurple", 13) == 0)) {
 			purple_debug_info(DBGID, "found plug-in %s\n", prpl_plugin->info->id);
-			purple_signal_connect(prpl_plugin, "twitter-message", plugin, PURPLE_CALLBACK(twitgin_on_tweet), NULL);
+			purple_signal_connect(prpl_plugin, "twitter-message", plugin, PURPLE_CALLBACK(twitgin_on_tweet_recv), NULL);
 		}
 	}
 
@@ -542,8 +755,8 @@ static gboolean plugin_unload(PurplePlugin *plugin)
 	purple_signal_disconnect(purple_get_core(), "uri-handler", plugin, PURPLE_CALLBACK(twittgin_uri_handler));
 #endif
 
-	purple_signal_disconnect(purple_conversations_get_handle(), "displaying-im-msg", plugin, PURPLE_CALLBACK(twitgin_on_displaying));
-	purple_signal_disconnect(pidgin_conversations_get_handle(), "twitgin-message", plugin, PURPLE_CALLBACK(twitgin_on_tweet));
+	purple_signal_disconnect(purple_conversations_get_handle(), "displaying-im-msg", plugin, PURPLE_CALLBACK(twitgin_on_tweet_send));
+	purple_signal_disconnect(pidgin_conversations_get_handle(), "twitgin-message", plugin, PURPLE_CALLBACK(twitgin_on_tweet_recv));
 
 	purple_debug_info(DBGID, "plugin unloaded\n");	
 	return TRUE;
